#include <sys/file.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

static int Error(int code, const char *info) {
  printf("RESULT\n");
  printf("info:%s\n", info);
  printf("code:%d\n", code);
  return code;
}

int main(int argv, char** argc) {

  struct stat st;

  int rfd, sfd;
  FILE *rf, *sf;
  fd_set fs;

  char buf[2048];
  char replyfile[] = "/tmp/giis-fifo.XXXXXX";

  struct timeval timelimit = { 5, 0 };

  const char* infile;

  signal(SIGPIPE, SIG_IGN);

  if(argv != 2)
    return Error(53, "Wrong number of arguments");

  infile = argc[1];

  if(stat(infile, &st) != 0)
    return Error(53, "Could not stat server FIFO");

  if(!S_ISFIFO(st.st_mode))
    return Error(53, "Server FIFO is not a FIFO");

  mktemp(replyfile);
  if(mkfifo(replyfile, S_IRUSR | S_IWUSR) == -1)
    return Error(53, "Could not create reply FIFO");

  sfd = open(infile, O_WRONLY | O_NONBLOCK);
  if (sfd == -1) {
    unlink(replyfile);
    return Error(53, "Could not open server FIFO");
  }
  if(flock(sfd, LOCK_EX) != 0) {
    close(sfd);
    unlink(replyfile);
    return Error(53, "Could not lock server FIFO");
  }
  sf = fdopen(sfd, "w");
  if (!sf) {
    flock(sfd, LOCK_UN);
    close(sfd);
    unlink(replyfile);
    return Error(53, "Could not open server FIFO");
  }

  rfd = open(replyfile, O_RDONLY | O_NONBLOCK);
  if (rfd == -1) {
    flock(sfd, LOCK_UN);
    fclose(sf);
    unlink(replyfile);
    return Error(53, "Could not open reply FIFO");
  }
  rf = fdopen(rfd, "r");
  if (!rf) {
    flock(sfd, LOCK_UN);
    fclose(sf);
    close(rfd);
    unlink(replyfile);
    return Error(53, "Could not open reply FIFO");
  }

  fprintf(sf, "%s\n", replyfile);

  while(fgets(buf, 2048, stdin)) {
    if(strncasecmp(buf, "timelimit:", 10) == 0)
      timelimit.tv_sec = atoi(&buf[10]);
    fputs(buf, sf);
  }

  fflush(sf);

  flock(sfd, LOCK_UN);
  fclose(sf);

  FD_ZERO(&fs);
  FD_SET(rfd, &fs);

  if(select(rfd + 1, &fs, NULL, NULL,
	    timelimit.tv_sec ? &timelimit : NULL) <= 0) {
    fclose(rf);
    unlink(replyfile);
    return Error(3, "Time out waiting for reply");
  }

  do {
    if(fgets(buf, 2048, rf))
      fputs(buf, stdout);
    else if(errno == EAGAIN)
      continue;
    else if(!feof(rf)) {
      fclose(rf);
      unlink(replyfile);
      return Error(53, "I/O Error while waiting for reply");
    }
  } while(!feof(rf));

  fclose(rf);
  unlink(replyfile);

  return 0;
}
