/* Quite an ugly hack to return all attributes from slapd-shell even if
   a attribute name is given.
   Needed in order to be compatible with clients expecting the non-standard
   behaviour of the Globus GIIS backend. */

#include <dlfcn.h>
#include <lber.h>
#include <ldap_features.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

typedef void** (*backend_info_t)(const char*);

typedef int (*main_t)(int, char**);

typedef struct arg_t {
  main_t main;
  int argv;
  char **argc;
} arg_t;

/* These are simplified versions of some structs and unions in the openldap
   internal header ldap.h
   Most pointers to structs have been replaced with void pointers, since
   only the size of the pointers is important. */

typedef struct req_search_s {
  int rs_scope;
  int rs_deref;
  int rs_slimit;
  int rs_tlimit;
  void *rs_limit;     /* struct slap_limits_set* */
  int rs_attrsonly;
  void *rs_attrs;     /* AttributeName*          */
  void *rs_filter;    /* Filter*                 */
  struct berval rs_filterstr;
} req_search_s;

#if LDAP_VENDOR_VERSION >= 20300  /* openldap 2.3 and later */

typedef union OpRequest {
  req_search_s oq_search;
  /* There are more fields in this union, but they are not relevant
     for this hack. */
} OpRequest;

typedef struct Operation {
  void *o_hdr;        /* Opheader*               */
  ber_tag_t o_tag;
  time_t o_time;
  int o_tincr;
  void *o_bd;         /* BackendDB*              */
  struct berval o_req_dn;
  struct berval o_req_ndn;
  OpRequest o_request;
  /* There are more fields in this struct, but they are not relevant
     for this hack. */
} Operation;

typedef struct SlapReply SlapReply;

#else  /* openldap 2.2 */

typedef struct slap_op {
  unsigned long o_opid;
  unsigned long o_connid;
  void *o_conn;       /* struct slap_conn*       */
  void *o_bd;         /* BackendDB*              */
  ber_int_t o_msgid;
  ber_int_t o_protocol;
  ber_tag_t o_tag;
  time_t o_time;
  struct berval o_req_dn;
  struct berval o_req_ndn;
  union o_req_u {
    req_search_s oq_search;
    /* There are more fields in this union, but they are not relevant
       for this hack. */
  } o_request;
  /* There are more fields in this struct, but they are not relevant
     for this hack. */
} Operation;

typedef struct slap_rep SlapReply;

#endif

typedef int (*BI_search_op)(Operation*, SlapReply*);

static BI_search_op shell_back_search;

static int shell_back_search_wrapper(Operation *op, SlapReply *rs) {
  void *save;
  int res;
  save = op->o_request.oq_search.rs_attrs;
  op->o_request.oq_search.rs_attrs = NULL;
  res = shell_back_search(op, rs);
  op->o_request.oq_search.rs_attrs = save;
  return res;
}

static void* main_wrapper(void *voidarg) {
  arg_t *arg = (arg_t*) voidarg;
  return (void*) arg->main(arg->argv, arg->argc);
}

int main(int argv, char **argc) {
  void *handle;
  backend_info_t backend_info;
  arg_t arg;
  void *res;
  void **backendInfo;
  pthread_t pid;
  int i;

  int daemonize = 1;

  for(i = 0; i < argv; i++) {
    if(strncmp(argc[i], "-d", 2) == 0) {
      daemonize = 0;
      break;
    } 
  }

  if(daemonize) {
    i = fork();
    if(i < 0)
      exit(1);
    if(i > 0)
      exit(0);
    setsid();
    if(freopen("/dev/null", "r", stdin) == NULL)
      fclose(stdin);
    if(freopen("/dev/null", "a", stdout) == NULL)
      fclose(stdout);
    if(freopen("/dev/null", "a", stderr) == NULL)
      fclose(stderr);
  }

  handle = dlopen("/usr/sbin/slapd", RTLD_LAZY);

  if(!handle) {
    fprintf(stderr, "Error: can't open /usr/sbin/slapd\n");
    exit(1);
  }

  backend_info = (backend_info_t) dlsym(handle, "backend_info");
  arg.main = (main_t) dlsym(handle, "main");
  if(daemonize) {
    arg.argv = argv + 2;
    arg.argc = (char**)malloc(sizeof(char*) * (argv + 2));
    for(i = 0; i < argv; i++)
      arg.argc[i] = argc[i];
    arg.argc[argv] = "-d";
    arg.argc[argv + 1] = "0";
  }
  else {
    arg.argv = argv;
    arg.argc = argc;
  }
  shell_back_search = (BI_search_op) dlsym(handle, "shell_back_search");

  if(!backend_info || !arg.main || !shell_back_search) {
    fprintf(stderr, "Error: missing symbols in slapd\n");
    exit(1);
  }

  pthread_create(&pid, NULL, &main_wrapper, (void*)&arg);

  i = 0;
  do {
    sleep(1); /* let the slapd server start */
    backendInfo = backend_info("shell");
    i++;
  }
  while (!backendInfo && i < 150);

  if(backendInfo)
    for(i = 0; i < 100; i++) {
      if(backendInfo[i] == shell_back_search) {
	backendInfo[i] = &shell_back_search_wrapper;
	break;
      }
    }
  else {
    fprintf(stderr, "Error: slapd shell backend not found - exiting\n");
    pthread_kill(pid, SIGTERM);
  }

  pthread_join(pid, &res);
  dlclose(handle);

  if(daemonize)
    free(arg.argc);

  return (int) res;
}
